"""liquidctl drivers for NZXT Smart Device V1/V2, Grid+ V3, HUE 2 and HUE 2 Ambient.

Smart Device (V1)
-----------------

The NZXT Smart Device is a fan and LED controller that ships with the H200i,
H400i, H500i and H700i cases.

It provides three independent fan channels with standard 4-pin connectors.
Both PWM and DC control is supported, and the device automatically chooses the
appropriate mode for each channel.

Additionally, up to four chained HUE+ LED strips, or five Aer RGB fans, can be
driven from only RGB channel available.  The firmware installed on the device
exposes several color presets, most of them common to other NZXT products.

The device recognizes the type of accessory connected by measuring the
resistance between the FD and GND lines.[1][2]  In normal usage accessories
should not be mixed.

A microphone is also present onboard, for noise level optimization through CAM
and AI.  NZXT calls this feature Adaptive Noise Reduction (ANR).

[1] https://forum.level1techs.com/t/nzxt-hue-a-look-inside/104836
[2] In parallel: 10 kOhm per HUE+ strip, 16 kOhm per Aer RGB fan.

Grid+ V3
--------

The NZXT Grid+ V3 is a fan controller very similar to the Smart Device (V1).
Comparing the two, the Grid+ has more fan channels (six in total), and no
support for LEDs.

Smart Device V2
---------------

The NZXT Smart Device V2 is a newer model of the original fan and LED controller. It
ships with NZXT's cases released in mid-2019 including the H510 Elite, H510i,
H710i, and H210i.

It provides three independent fan channels with standard 4-pin connectors. Both
PWM and DC control is supported, and the device automatically chooses the appropriate
mode for each channel.

Additionally, it features two independent lighting (Addressable RGB) channels,
unlike the single channel in the original. NZXT Aer RGB 2 fans and HUE 2 lighting
accessories (HUE 2 LED strip, HUE 2 Unerglow, HUE 2 Cable Comb) can be
connected. The firmware installed on the device exposes several color presets, most
of them common to other NZXT products.

HUE 2 and HUE+ devices (including Aer RGB and Aer RGB 2 fans) are supported, but
HUE 2 components cannot be mixed with HUE+ components in the same channel. Each
lighting channel supports up to 6 accessories and a total of 40 LEDs.

A microphone is still present onboard for noise level optimization through CAM
and AI.

NZXT H1 V2
----------

The second revision of the NZXT H1 case, labeled H1 V2, ships with a variant of the
NZXT Smart Device V2 that handles both the internal fans and the AIO pump. Two fan and
one pump channels are available, where the formers can be controlled via PWM or DC.

The pump speed is not user controllable. The device reports the state, speed and duty
of each fan channel, as well as the pump speed.

There are no lighting channels available nor an onboard microphone.

RGB & Fan Controller
--------------------

The NZXT RGB & Fan Controller is a retail version of the NZXT Smart Device V2.

HUE 2
-----

The NZXT HUE 2 is an LED controller from the same generation of the Smart
Device V2.

The presets and limitations of the four LED channels are the same as in the
Smart Device V2.

HUE 2 Ambient
-------------

HUE 2 Ambient is a variant of HUE 2 featuring 2 LED control channels.

Driver
------

This driver implements all features available at the hardware level:

 - initialization
 - detection of connected fans and LED strips
 - control of fan speeds per channel
 - monitoring of fan presence, control mode, speed, voltage and current
 - control of lighting modes and colors
 - reporting of LED accessory count and type
 - monitoring of noise level (from the onboard microphone)
 - reporting of firmware version

Software based features offered by CAM, like ANR, have not been implemented.

After powering on from Mechanical Off, or if there have been hardware changes,
the devices must be manually initialized by calling `initialize()`.  This will
cause all connected fans and LED accessories to be detected, and enable status
updates.  It is recommended to initialize the devices at every boot.

Copyright Jonas Malaco, CaseySJ and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import itertools
import logging
import time

from liquidctl.driver.usb import UsbHidDriver
from liquidctl.error import NotSupportedByDevice
from liquidctl.util import clamp, map_direction, Hue2Accessory, \
                           HUE2_MAX_ACCESSORIES_IN_CHANNEL

_LOGGER = logging.getLogger(__name__)

_ANIMATION_SPEEDS = {
    'slowest':  0x0,
    'slower':   0x1,
    'normal':   0x2,
    'faster':   0x3,
    'fastest':  0x4,
}

_MIN_DUTY = 0
_MAX_DUTY = 100


class _BaseSmartDevice(UsbHidDriver):
    """Common functions of Smart Device and Grid drivers."""

    def __init__(self, device, description, speed_channels, color_channels, **kwargs):
        super().__init__(device, description)
        self._speed_channels = speed_channels
        self._color_channels = color_channels

    def set_color(self, channel, mode, colors, speed='normal', direction='forward', **kwargs):
        """Set the color mode for a specific channel.

        Only supported by Smart Device V1/V2 and HUE 2 controllers.
        """

        if not self._color_channels:
            raise NotSupportedByDevice()

        if 'backwards' in mode:
            _LOGGER.warning('deprecated mode, move to direction=backward option')
            mode = mode.replace('backwards-', '')
            direction = 'backward'

        cid = self._color_channels[channel]
        _, _, _, mincolors, maxcolors = self._COLOR_MODES[mode]
        colors = [[g, r, b] for [r, g, b] in colors]
        if len(colors) < mincolors:
            raise ValueError(f'not enough colors for mode={mode}, at least {mincolors} required')
        elif maxcolors == 0:
            if colors:
                _LOGGER.warning('too many colors for mode=%s, none needed', mode)
            colors = [[0, 0, 0]]  # discard the input but ensure at least one step
        elif len(colors) > maxcolors:
            _LOGGER.warning('too many colors for mode=%s, dropping to %d',
                            mode, maxcolors)
            colors = colors[:maxcolors]

        sval = _ANIMATION_SPEEDS[speed]
        self._write_colors(cid, mode, colors, sval, direction)

    def set_fixed_speed(self, channel, duty, **kwargs):
        """Set channel to a fixed speed duty."""

        if channel == 'sync':
            selected_channels = self._speed_channels
        else:
            selected_channels = {channel: self._speed_channels[channel]}
        for cname, (cid, dmin, dmax) in selected_channels.items():
            duty = clamp(duty, dmin, dmax)
            _LOGGER.info('setting %s duty to %d%%', cname, duty)
            self._write_fixed_duty(cid, duty)

    def set_speed_profile(self, channel, profile, **kwargs):
        raise NotSupportedByDevice()

    def _write(self, data):
        padding = [0x0]*(self._WRITE_LENGTH - len(data))
        self.device.write(data + padding)

    def _write_colors(self, cid, mode, colors, sval, direction):
        raise NotImplementedError()

    def _write_fixed_duty(self, cid, duty):
        raise NotImplementedError()

    def set_screen(self, channel, mode, value, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()


class SmartDevice(_BaseSmartDevice):
    """NZXT Smart Device (V1) or Grid+ V3."""

    # support for hwmon: nzxt-grid3, liquidtux
    # https://github.com/liquidctl/liquidtux/blob/3b80dafead6f/nzxt-grid3.c

    _MATCHES = [
        (0x1e71, 0x1714, 'NZXT Smart Device (V1)', {
            'speed_channel_count': 3,
            'color_channel_count': 1
        }),
        (0x1e71, 0x1711, 'NZXT Grid+ V3', {
            'speed_channel_count': 6,
            'color_channel_count': 0
        }),
    ]

    _READ_LENGTH = 21
    _WRITE_LENGTH = 65

    _COLOR_MODES = {
        # (byte2/mode, byte3/variant, byte4/size, min colors, max colors)
        'off':                           (0x00, 0x00, 0x00, 0, 0),
        'fixed':                         (0x00, 0x00, 0x00, 1, 1),
        'super-fixed':                   (0x00, 0x00, 0x00, 1, 40),  # independent leds
        'fading':                        (0x01, 0x00, 0x00, 1, 8),
        'spectrum-wave':                 (0x02, 0x00, 0x00, 0, 0),
        'marquee-3':                     (0x03, 0x00, 0x00, 1, 1),
        'marquee-4':                     (0x03, 0x00, 0x08, 1, 1),
        'marquee-5':                     (0x03, 0x00, 0x10, 1, 1),
        'marquee-6':                     (0x03, 0x00, 0x18, 1, 1),
        'covering-marquee':              (0x04, 0x00, 0x00, 1, 8),
        'alternating':                   (0x05, 0x00, 0x00, 2, 2),
        'moving-alternating':            (0x05, 0x08, 0x00, 2, 2),
        'pulse':                         (0x06, 0x00, 0x00, 1, 8),
        'breathing':                     (0x07, 0x00, 0x00, 1, 8),   # colors for each step
        'super-breathing':               (0x07, 0x00, 0x00, 1, 40),  # one step, independent leds
        'candle':                        (0x09, 0x00, 0x00, 1, 1),
        'wings':                         (0x0c, 0x00, 0x00, 1, 1),
        'super-wave':                    (0x0d, 0x00, 0x00, 1, 40),  # independent ring leds

        # deprecated in favor of direction=backward
        'backwards-spectrum-wave':       (0x02, 0x00, 0x00, 0, 0),
        'backwards-marquee-3':           (0x03, 0x00, 0x00, 1, 1),
        'backwards-marquee-4':           (0x03, 0x00, 0x08, 1, 1),
        'backwards-marquee-5':           (0x03, 0x00, 0x10, 1, 1),
        'backwards-marquee-6':           (0x03, 0x00, 0x18, 1, 1),
        'covering-backwards-marquee':    (0x04, 0x00, 0x00, 1, 8),
        'backwards-moving-alternating':  (0x05, 0x08, 0x00, 2, 2),
        'backwards-super-wave':          (0x0d, 0x00, 0x00, 1, 40),
    }

    def __init__(self, device, description, speed_channel_count, color_channel_count, **kwargs):
        """Instantiate a driver with a device handle."""
        speed_channels = {f'fan{i + 1}': (i, _MIN_DUTY, _MAX_DUTY)
                          for i in range(speed_channel_count)}
        color_channels = {'led': (i)
                          for i in range(color_channel_count)}
        super().__init__(device, description, speed_channels, color_channels, **kwargs)

    def initialize(self, direct_access=False, **kwargs):
        """Initialize the device and the driver.

        Connected fans and LED accessories are detected.

        This method should be called every time the systems boots, resumes from
        a suspended state, or if the device has just been (re)connected.  In
        those scenarios, no other method, except `connect()` or `disconnect()`,
        should be called until the device and driver has been (re-)initialized.

        Returns None or a list of `(property, value, unit)` tuples, similarly
        to `get_status()`.
        """

        if self._hwmon and not direct_access:
            _LOGGER.info('bound to %s kernel driver, assuming it is already initialized',
                         self._hwmon.driver)
        else:
            if self._hwmon:
                _LOGGER.warning('forcing re-initialization despite %s kernel driver',
                                self._hwmon.driver)
            self._write([0x1, 0x5c])  # initialize/detect connected devices and their type
            self._write([0x1, 0x5d])  # start reporting
            self.device.clear_enqueued_reports()

        msg = self.device.read(self._READ_LENGTH)
        fw = tuple(msg[0xb:0xf])
        _LOGGER.debug('raw firmware version: %r', fw)

        # NZXT/CAM has simplified how they report the firmware version from its
        # raw 4-component form
        fw_human = f'{fw[0]}.{fw[3]}'
        ret = [('Firmware version', fw_human, '')]

        if self._color_channels:
            lcount = msg[0x11]
            ret.append(('LED accessories', lcount, ''))
            if lcount > 0:
                ltype, lsize = [('HUE+ Strip', 10), ('Aer RGB', 8)][msg[0x10] >> 3]
                ret.append(('LED accessory type', ltype, ''))
                ret.append(('LED count (total)', lcount*lsize, ''))

        return ret

    def _get_status_directly(self):
        fans = [None] * len(self._speed_channels)
        noise = []

        self.device.clear_enqueued_reports()
        for i, _ in enumerate(fans):
            msg = self.device.read(self._READ_LENGTH)
            num = (msg[15] >> 4) + 1
            state = msg[15] & 0x3

            fans[num - 1] = [
                (f'Fan {num} speed', msg[3] << 8 | msg[4], 'rpm'),
                (f'Fan {num} voltage', msg[7] + msg[8]/100, 'V'),
                (f'Fan {num} current', msg[9] + msg[10]/100, 'A'),
                (f'Fan {num} control mode', [None, 'DC', 'PWM'][state], ''),
            ]
            noise.append(msg[1])

        # flatten fan data while checking for holes
        ret = []
        for i, fan in enumerate(fans):
            if fan:
                ret = ret + fan
            else:
                _LOGGER.warning('missing data fan for %d', i + 1)

        ret.append(('Noise level', round(sum(noise)/len(noise)), 'dB'))

        return ret

    def _get_status_from_hwmon(self):
        ret = []
        mode = ['DC', 'PWM']  # slightly simplified, but the device treats undetected == PWM

        for i in range(len(self._speed_channels)):
            n = i + 1
            ret.append((f'Fan {n} speed', self._hwmon.read_int(f'fan{n}_input'), 'rpm')),
            ret.append((f'Fan {n} voltage', self._hwmon.read_int(f'in{i}_input') * 1e-3, 'V')),
            ret.append((f'Fan {n} current', self._hwmon.read_int(f'curr{n}_input') * 1e-3, 'A')),
            ret.append((f'Fan {n} control mode', mode[self._hwmon.read_int(f'pwm{n}_mode')], '')),

        # noise level is not available through hwmon, but also not very accurate or useful

        return ret

    def get_status(self, direct_access=False, **kwargs):
        """Get a status report.

        Returns a list of `(property, value, unit)` tuples.
        """

        if self._hwmon and not direct_access:
            _LOGGER.info('bound to %s kernel driver, reading status from hwmon', self._hwmon.driver)
            return self._get_status_from_hwmon()

        if self._hwmon:
            _LOGGER.warning('directly reading the status despite %s kernel driver',
                            self._hwmon.driver)

        return self._get_status_directly()

    def _write_colors(self, cid, mode, colors, sval, direction='forward'):
        mval, mod3, mod4, _, _ = self._COLOR_MODES[mode]
        # generate steps from mode and colors: usually each color set by the user generates
        # one step, where it is specified to all leds and the device handles the animation;
        # but in super mode there is a single step and each color directly controls a led

        mod3 += map_direction(direction, 0, 0x10)

        if 'super' in mode:
            steps = [list(itertools.chain(*colors))]
        else:
            steps = [color*40 for color in colors]
        for i, leds in enumerate(steps):
            seq = i << 5
            byte4 = sval | seq | mod4
            self._write([0x2, 0x4b, mval, mod3, byte4] + leds[0:57])
            self._write([0x3] + leds[57:])

    def _write_fixed_duty(self, cid, duty):
        self._write([0x2, 0x4d, cid, 0, duty])


class SmartDevice2(_BaseSmartDevice):
    """NZXT HUE 2 lighting and, optionally, fan controller."""

    # support for hwmon: nzxt-smart2, Linux 5.17

    _MATCHES = [
        (0x1e71, 0x2006, 'NZXT Smart Device V2', {
            'speed_channel_count': 3,
            'color_channel_count': 2
        }),
        (0x1e71, 0x200d, 'NZXT Smart Device V2', {
            'speed_channel_count': 3,
            'color_channel_count': 2
        }),
        (0x1e71, 0x200f, 'NZXT Smart Device V2', {
            'speed_channel_count': 3,
            'color_channel_count': 2
        }),
        (0x1e71, 0x2001, 'NZXT HUE 2', {
            'speed_channel_count': 0,
            'color_channel_count': 4
        }),
        (0x1e71, 0x2002, 'NZXT HUE 2 Ambient', {
            'speed_channel_count': 0,
            'color_channel_count': 2
        }),
        (0x1e71, 0x2009, 'NZXT RGB & Fan Controller', {
            'speed_channel_count': 3,
            'color_channel_count': 2
        }),
        (0x1e71, 0x200e, 'NZXT RGB & Fan Controller', {
            'speed_channel_count': 3,
            'color_channel_count': 2
        }),
        (0x1e71, 0x2010, 'NZXT RGB & Fan Controller', {
            'speed_channel_count': 3,
            'color_channel_count': 2
        }),
        (0x1e71, 0x2011, 'NZXT RGB & Fan Controller (3+6 channels) (experimental)', {
            'speed_channel_count': 3,
            'color_channel_count': 0
        }),
        (0x1e71, 0x2019, 'NZXT RGB & Fan Controller (3+6 channels) (experimental)', {
            'speed_channel_count': 3,
            'color_channel_count': 0
        }),
    ]

    _MAX_READ_ATTEMPTS = 12
    _READ_LENGTH = 64
    _WRITE_LENGTH = 64

    _COLOR_MODES = {
        # (mode, size/variant, moving, min colors, max colors)
        'off':                              (0x00, 0x00, 0x00, 0, 0),
        'fixed':                            (0x00, 0x00, 0x00, 1, 1),
        'super-fixed':                      (0x01, 0x00, 0x00, 1, 40),  # independent leds
        'fading':                           (0x01, 0x00, 0x00, 1, 8),
        'spectrum-wave':                    (0x02, 0x00, 0x00, 0, 0),
        'marquee-3':                        (0x03, 0x00, 0x00, 1, 1),
        'marquee-4':                        (0x03, 0x01, 0x00, 1, 1),
        'marquee-5':                        (0x03, 0x02, 0x00, 1, 1),
        'marquee-6':                        (0x03, 0x03, 0x00, 1, 1),
        'covering-marquee':                 (0x04, 0x00, 0x00, 1, 8),
        'alternating-3':                    (0x05, 0x00, 0x00, 2, 2),
        'alternating-4':                    (0x05, 0x01, 0x00, 2, 2),
        'alternating-5':                    (0x05, 0x02, 0x00, 2, 2),
        'alternating-6':                    (0x05, 0x03, 0x00, 2, 2),
        'moving-alternating-3':             (0x05, 0x00, 0x10, 2, 2),   # byte4: 0x10 = moving
        'moving-alternating-4':             (0x05, 0x01, 0x10, 2, 2),   # byte4: 0x10 = moving
        'moving-alternating-5':             (0x05, 0x02, 0x10, 2, 2),   # byte4: 0x10 = moving
        'moving-alternating-6':             (0x05, 0x03, 0x10, 2, 2),   # byte4: 0x10 = moving
        'pulse':                            (0x06, 0x00, 0x00, 1, 8),
        'breathing':                        (0x07, 0x00, 0x00, 1, 8),   # colors for each step
        'super-breathing':                  (0x03, 0x19, 0x00, 1, 40),  # independent leds
        'candle':                           (0x08, 0x00, 0x00, 1, 1),
        'starry-night':                     (0x09, 0x00, 0x00, 1, 1),
        'rainbow-flow':                     (0x0b, 0x00, 0x00, 0, 0),
        'super-rainbow':                    (0x0c, 0x00, 0x00, 0, 0),
        'rainbow-pulse':                    (0x0d, 0x00, 0x00, 0, 0),
        'wings':                            (None, 0x00, 0x00, 1, 1),   # wings requires special handling

        # deprecated in favor of direction=backward
        'backwards-spectrum-wave':          (0x02, 0x00, 0x00, 0, 0),
        'backwards-marquee-3':              (0x03, 0x00, 0x00, 1, 1),
        'backwards-marquee-4':              (0x03, 0x01, 0x00, 1, 1),
        'backwards-marquee-5':              (0x03, 0x02, 0x00, 1, 1),
        'backwards-marquee-6':              (0x03, 0x03, 0x00, 1, 1),
        'covering-backwards-marquee':       (0x04, 0x00, 0x00, 1, 8),
        'backwards-moving-alternating-3':   (0x05, 0x00, 0x01, 2, 2),
        'backwards-moving-alternating-4':   (0x05, 0x01, 0x01, 2, 2),
        'backwards-moving-alternating-5':   (0x05, 0x02, 0x01, 2, 2),
        'backwards-moving-alternating-6':   (0x05, 0x03, 0x01, 2, 2),
        'backwards-rainbow-flow':           (0x0b, 0x00, 0x00, 0, 0),
        'backwards-super-rainbow':          (0x0c, 0x00, 0x00, 0, 0),
        'backwards-rainbow-pulse':          (0x0d, 0x00, 0x00, 0, 0),
    }

    def __init__(self, device, description, speed_channel_count, color_channel_count, **kwargs):
        """Instantiate a driver with a device handle."""
        speed_channels = {f'fan{i + 1}': (i, _MIN_DUTY, _MAX_DUTY)
                          for i in range(speed_channel_count)}
        color_channels = {f'led{i + 1}': (1 << i)
                          for i in range(color_channel_count)}
        if color_channels:
            color_channels['sync'] = (1 << color_channel_count) - 1
        super().__init__(device, description, speed_channels, color_channels, **kwargs)

    def initialize(self, direct_access=False, **kwargs):
        """Initialize the device and the driver.

        Connected fans and LED accessories are detected.

        This method should be called every time the systems boots, resumes from
        a suspended state, or if the device has just been (re)connected.  In
        those scenarios, no other method, except `connect()` or `disconnect()`,
        should be called until the device and driver has been (re-)initialized.

        Returns None or a list of `(property, value, unit)` tuples, similarly
        to `get_status()`.
        """

        self.device.clear_enqueued_reports()

        # if fan controller, initialize fan reporting (#331)
        if self._speed_channels:
            if self._hwmon and not direct_access:
                _LOGGER.info('bound to %s kernel driver, assuming it is already initialized',
                             self._hwmon.driver)
            else:
                if self._hwmon:
                    _LOGGER.warning('forcing re-initialization despite %s kernel driver',
                                    self._hwmon.driver)
                update_interval = (lambda secs: 1 + round((secs - .5) / .25))(.5)  # see issue #128
                self._write([0x60, 0x02, 0x01, 0xe8, update_interval, 0x01, 0xe8, update_interval])
                self._write([0x60, 0x03])

        # request static infos
        self._write([0x10, 0x01])  # firmware info
        self._write([0x20, 0x03])  # lighting info
        ret = []

        def parse_firm_info(msg):
            fw = f'{msg[0x11]}.{msg[0x12]}.{msg[0x13]}'
            ret.append(('Firmware version', fw, ''))

        def parse_led_info(msg):
            channel_count = msg[14]
            offset = 15  # offset of first channel/first accessory
            for c in range(channel_count):
                for a in range(HUE2_MAX_ACCESSORIES_IN_CHANNEL):
                    accessory_id = msg[offset + c * HUE2_MAX_ACCESSORIES_IN_CHANNEL + a]
                    if accessory_id == 0:
                        break
                    ret.append((f'LED {c + 1} accessory {a + 1}',
                                   Hue2Accessory(accessory_id), ''))

        parsers = {b'\x11\x01': parse_firm_info}
        if self._color_channels:
            parsers[b'\x21\x03'] = parse_led_info
        self._read_until(parsers)
        return sorted(ret)

    def _get_status_directly(self):
        ret = []

        def parse_fan_info(msg):
            mode_offset = 16
            rpm_offset = 24
            duty_offset = 40
            noise_offset = 56
            raw_modes = [None, 'DC', 'PWM']

            for i, _ in enumerate(self._speed_channels):
                mode = raw_modes[msg[mode_offset + i]]
                ret.append((f'Fan {i + 1} speed', msg[rpm_offset + 1] << 8 | msg[rpm_offset], 'rpm'))
                ret.append((f'Fan {i + 1} duty', msg[duty_offset + i], '%'))
                ret.append((f'Fan {i + 1} control mode', mode, ''))
                rpm_offset += 2
            ret.append(('Noise level', msg[noise_offset], 'dB'))

        self.device.clear_enqueued_reports()
        self._read_until({b'\x67\x02': parse_fan_info})
        return sorted(ret)

    def _get_status_from_hwmon(self):
        ret = []
        modes = ['DC', 'PWM']

        for n in range(1, len(self._speed_channels) + 1):
            ret.append((f'Fan {n} speed', self._hwmon.read_int(f'fan{n}_input'), 'rpm')),
            ret.append((f'Fan {n} duty', self._hwmon.read_int(f'pwm{n}') * 100. / 255, '%')),
            ret.append((f'Fan {n} control mode', modes[self._hwmon.read_int(f'pwm{n}_mode')], '')),

        # noise level is not available through hwmon, but also not very accurate or useful

        return sorted(ret)

    def get_status(self, direct_access=False, **kwargs):
        """Get a status report.

        Returns a list of `(property, value, unit)` tuples.
        """

        if not self._speed_channels:
            return []

        if self._hwmon and not direct_access:
            _LOGGER.info('bound to %s kernel driver, reading status from hwmon', self._hwmon.driver)
            return self._get_status_from_hwmon()

        if self._hwmon:
            _LOGGER.warning('directly reading the status despite %s kernel driver',
                            self._hwmon.driver)

        return self._get_status_directly()

    def _read_until(self, parsers):
        for _ in range(self._MAX_READ_ATTEMPTS):
            msg = self.device.read(self._READ_LENGTH)
            prefix = bytes(msg[0:2])
            func = parsers.pop(prefix, None)
            if func:
                func(msg)
            if not parsers:
                return
        assert False, f'missing messages (attempts={self._MAX_READ_ATTEMPTS}, missing={len(parsers)})'

    def _write_colors(self, cid, mode, colors, sval, direction='forward',):
        mval, mod3, mod4, mincolors, maxcolors = self._COLOR_MODES[mode]

        if not self._color_channels:
            raise NotSupportedByDevice()

        color_count = len(colors)
        if maxcolors == 40:
            led_padding = [0x00, 0x00, 0x00]*(maxcolors - color_count)  # turn off remaining LEDs
            leds = list(itertools.chain(*colors)) + led_padding
            self._write([0x22, 0x10, cid, 0x00] + leds[0:60])  # send first 20 colors to device (3 bytes per color)
            self._write([0x22, 0x11, cid, 0x00] + leds[60:])  # send remaining colors to device
            self._write([0x22, 0xa0, cid, 0x00, mval, mod3, 0x00, 0x00, 0x00,
                         0x00, 0x64, 0x00, 0x32, 0x00, 0x00, 0x01])
        elif mode == 'wings':  # wings requires special handling
            for [g, r, b] in colors:
                self._write([0x22, 0x10, cid])  # clear out all independent LEDs
                self._write([0x22, 0x11, cid])  # clear out all independent LEDs
                color_lists = [] * 3
                color_lists[0] = [g, r, b] * 8
                color_lists[1] = [int(x // 2.5) for x in color_lists[0]]
                color_lists[2] = [int(x // 4) for x in color_lists[1]]
                for i in range(8):   # send color scheme first, before enabling wings mode
                    mod = 0x05 if i in [3, 7] else 0x01
                    msg = ([0x22, 0x20, cid, i, 0x04, 0x39, 0x00, mod,
                            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06,
                            0x05, 0x85, 0x05, 0x85, 0x05, 0x85, 0x00, 0x00,
                            0x00, 0x00, 0x00, 0x00])
                    self._write(msg + color_lists[i % 4])
                self._write([0x22, 0x03, cid, 0x08])   # this actually enables wings mode
        else:
            byte7 = (mod4 & 0x10) >> 4  # sets 'moving' flag for moving alternating modes
            byte8 = map_direction(direction, 0, 1)  # sets 'backward' flag
            byte9 = mod3 if mval == 0x03 else color_count  # specifies 'marquee' LED size
            byte10 = mod3 if mval == 0x05 else 0x00  # specifies LED size for 'alternating' modes
            header = [0x28, 0x03, cid, 0x00, mval, sval, byte7, byte8, byte9, byte10]
            self._write(header + list(itertools.chain(*colors)))

    def _write_fixed_duty(self, cid, duty):
        msg = [0x62, 0x01, 0x01 << cid, 0x00, 0x00, 0x00]  # fan channel passed as bitflag in last 3 bits of 3rd byte
        msg[cid + 3] = duty  # duty percent in 4th, 5th, and 6th bytes for, respectively, fan1, fan2 and fan3
        self._write(msg)


class H1V2(SmartDevice2):
    _MATCHES = [
        (0x1e71, 0x2015, 'NZXT H1 V2 (experimental)', {
            'speed_channel_count': 2,
            'color_channel_count': 0
        }),
    ]

    def get_status(self, direct_access=False, **kwargs):
        ret = []

        def parse_fan_info(msg):
            mode_offset = 21
            rpm_offset = 24
            duty_offset = 25
            pump_offset = 18
            raw_modes = [None, 'DC', 'PWM']

            for i, _ in enumerate(self._speed_channels):
                mode = raw_modes[msg[mode_offset + i]]
                ret.append((f'Fan {i + 1} speed', msg[rpm_offset] << 8 | msg[rpm_offset - 1], 'rpm'))
                ret.append((f'Fan {i + 1} duty', msg[duty_offset], '%'))
                ret.append((f'Fan {i + 1} control mode', mode, ''))
                rpm_offset += 5
                duty_offset += 5
            ret.append(('Pump speed', msg[pump_offset] << 8 | msg[pump_offset - 1], 'rpm'))

        # parse fans and pump status
        self.device.clear_enqueued_reports()
        self._read_until({b'\x75\x02': parse_fan_info})
        return sorted(ret)


# backward compatibility
NzxtSmartDeviceDriver = SmartDevice
SmartDeviceDriver = SmartDevice
SmartDeviceV2Driver = SmartDevice2



#!/usr/bin/env python
# Copyright (C) 2022–2022  Peter Eckersley <pde@pde.is>
# SPDX-License-Identifier: GPL-3.0-or-later

import argparse
import sys
import time

import liquidctl.cli as lc
import liquidctl


try:
    import coloraide as c
except ImportError:
    print("The python coloraide package is not installed...")
    print("Try installing it with: pip install coloraide")
    sys.exit(1)


teal = c.Color("#2de544")
bluer = c.Color("#2d8a8d")
deep_blue = c.Color("#001030")
purple = c.Color("#c000c0")
red = c.Color("#d00000")

# set up argparse
parser = argparse.ArgumentParser(
    description="Cycle through colors on an Gigabyte RGB Fusion 2.0 device."
)
parser.add_argument(
    "--space",
    type=str,
    default="srgb",
    help="Color space to use; see https://facelessuser.github.io/coloraide/colors/",
)
# you can provide 0 or more colors to cycle through
parser.add_argument("colors", nargs="*", help="Colors to cycle through")
parser.add_argument(
    "--steps", default=30, type=int, help="Number of steps per color fade (default: 30)"
)
parser.add_argument("--debug", action="store_true", help="Print debug messages")
parser.add_argument(
    "--channel",
    type=str,
    default="led6",
    help="Channel to use; see https://github.com/liquidctl/liquidctl/blob/main/docs/gigabyte-rgb-fusion2-guide.md",
)
parser.add_argument(
    "--speed", type=float, default=1.0, help="Speed (seconds per color transition)"
)

args = parser.parse_args()

colors = [c.Color("#" + x) for x in args.colors]
if len(colors) < 2:
    colors.append(teal)
if len(colors) < 2:
    colors.append(bluer)

devs = liquidctl.driver.rgb_fusion2.RgbFusion2.find_supported_devices()
if not devs:
    print("No Gigabyte RGB Fusion 2.0 device found.")
    sys.exit(1)
elif len(devs) > 1:
    print("Multiple Gigabyte RGB Fusion 2.0 devices found, using first one.")

dev = devs[0]

lookup = []


def to_int(color):
    string = color.to_string()  # "rgb(255 75 0)"
    string = string.partition("(")[2][:-1]  # "255 75 0"
    string = string.split()  # ["255", "75", "0"]
    return [int(float(x)) for x in string]


colors.append(colors[0])  # add first color at the end, so we can loop

for i in range(len(colors) - 1):
    gradient = colors[i].interpolate(colors[i + 1], space=args.space)
    cols = [to_int(gradient(x / args.steps)) for x in range(args.steps)]
    lookup += cols

if args.debug:
    print("Cycling through", len(lookup), "colors:\n", lookup)


while True:
    # reconnect occasionally, just in case these connections die and we can bring them back
    with dev.connect():
        for c in lookup:
            dev.set_color(channel=args.channel, mode="fixed", colors=[c])
            time.sleep(args.speed * 2.0 / args.steps)



"""Test backward compatibility with liquidctl 1.3.x."""

import pytest


def test_pre14_old_module_names():
    import liquidctl.driver.asetek
    import liquidctl.driver.corsair_hid_psu
    import liquidctl.driver.kraken_two
    import liquidctl.driver.nzxt_smart_device
    import liquidctl.driver.seasonic


def test_pre14_old_driver_names():
    from liquidctl.driver.asetek import AsetekDriver
    from liquidctl.driver.asetek import LegacyAsetekDriver
    from liquidctl.driver.asetek import CorsairAsetekDriver
    from liquidctl.driver.corsair_hid_psu import CorsairHidPsuDriver
    from liquidctl.driver.kraken_two import KrakenTwoDriver
    from liquidctl.driver.nzxt_smart_device import SmartDeviceDriver
    from liquidctl.driver.nzxt_smart_device import SmartDeviceV2Driver
    from liquidctl.driver.seasonic import SeasonicEDriver



# uses the psf/black style

import pytest
from _testutils import MockHidapiDevice, Report

from liquidctl.driver.hwmon import HwmonDevice
from liquidctl.driver.smart_device import H1V2

SAMPLE_STATUS = bytes.fromhex(
    "75021320020d85bcabab94188f5f010000a00f0032020284021e1e02f9066464"
    "0000000000000000000000000000000000000000000000000000000000000005"
)


class MockH1V2(MockHidapiDevice):
    def __init__(self, raw_speed_channels, raw_led_channels):
        super().__init__()
        self.raw_speed_channels = raw_speed_channels
        self.raw_led_channels = raw_led_channels

    def write(self, data):
        reply = bytearray(64)
        if data[0:2] == [0x10, 0x01]:
            reply[0:2] = [0x11, 0x01]
        elif data[0:2] == [0x20, 0x03]:
            reply[0:2] = [0x21, 0x03]
            reply[14] = self.raw_led_channels
            if self.raw_led_channels > 1:
                reply[15 + 1 * 6] = 0x10
                reply[15 + 2 * 6] = 0x11
        self.preload_read(Report(reply[0], reply[1:]))
        return super().write(data)


@pytest.fixture
def mock_smart2():
    raw = MockH1V2(raw_speed_channels=2, raw_led_channels=0)
    dev = H1V2(raw, "Mock H1 V2", speed_channel_count=2, color_channel_count=0)
    dev.connect()
    return dev


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True), (True, False)])
def test_initializes(mock_smart2, has_hwmon, direct_access, tmp_path):
    if has_hwmon:
        mock_smart2._hwmon = HwmonDevice("mock_module", tmp_path)

    _ = mock_smart2.initialize(direct_access=direct_access)

    writes = len(mock_smart2.device.sent)
    if not has_hwmon or direct_access:
        assert writes == 4
    else:
        assert writes == 2


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (False, True), (True, True)])
def test_reads_status(mock_smart2, has_hwmon, direct_access):
    if has_hwmon:
        mock_smart2._hwmon = HwmonDevice(None, None)

    mock_smart2.device.preload_read(Report(0, SAMPLE_STATUS))

    expected = [
        ("Fan 1 control mode", "PWM", ""),
        ("Fan 1 duty", 30, "%"),
        ("Fan 1 speed", 644, "rpm"),
        ("Fan 2 control mode", "PWM", ""),
        ("Fan 2 duty", 100, "%"),
        ("Fan 2 speed", 1785, "rpm"),
        ("Pump speed", 4000, "rpm"),
    ]

    got = mock_smart2.get_status(direct_access=direct_access)

    assert sorted(got) == sorted(expected)


def test_constructor_sets_up_all_channels(mock_smart2):
    assert mock_smart2._speed_channels == {
        "fan1": (0, 0, 100),
        "fan2": (1, 0, 100),
    }


def test_not_totally_broken(mock_smart2):
    _ = mock_smart2.initialize()
    mock_smart2.device.preload_read(Report(0, [0x75, 0x02] + [0] * 62))
    _ = mock_smart2.get_status()
    mock_smart2.set_fixed_speed(channel="fan2", duty=50)



"""liquidctl – monitor and control liquid coolers and other devices.

Usage:
  liquidctl [options] list
  liquidctl [options] initialize [all]
  liquidctl [options] status
  liquidctl [options] set <channel> speed (<temperature> <percentage>) ...
  liquidctl [options] set <channel> speed <percentage>
  liquidctl [options] set <channel> color <mode> [<color>] ...
  liquidctl [options] set <channel> screen <mode> [<value>]
  liquidctl --help
  liquidctl --version

Device selection options (see: list -v):
  -m, --match <substring>            Filter devices by description substring
  -n, --pick <number>                Pick among many results for a given filter
  --vendor <id>                      Filter devices by hexadecimal vendor ID
  --product <id>                     Filter devices by hexadecimal product ID
  --release <number>                 Filter devices by hexadecimal release number
  --serial <number>                  Filter devices by serial number
  --bus <bus>                        Filter devices by bus
  --address <address>                Filter devices by address in bus
  --usb-port <port>                  Filter devices by USB port in bus

Animation options (devices/modes can support zero or more):
  --speed <value>                    Abstract animation speed (device/mode specific)
  --time-per-color <value>           Time to wait on each color (seconds)
  --time-off <value>                 Time to wait with the LED turned off (seconds)
  --alert-threshold <number>         Threshold temperature for a visual alert (°C)
  --alert-color <color>              Color used by the visual high temperature alert
  --direction <string>               If the pattern should move forward or backward.
  --start-led <number>               The first led to start the effect at
  --maximum-leds <number>            The number of LED's the effect should apply to

Other device options:
  --single-12v-ocp                   Enable single rail +12V OCP
  --pump-mode <mode>                 Set the pump mode (certain Corsair coolers)
  --fan-mode <channel>:<mode>[,...]  Set the mode for each fan (certain Corsair devices)
  --temperature-sensor <number>      The temperature sensor number for the Commander Pro
  --legacy-690lc                     Use Asetek 690LC in legacy mode (old Krakens)
  --non-volatile                     Store on non-volatile controller memory
  --direct-access                    Directly access the device despite kernel drivers
  --unsafe <features>                Comma-separated bleeding-edge features to enable

Other interface options:
  -v, --verbose                      Output additional information
  -g, --debug                        Show debug information on stderr
  --json                             JSON output (list/initialization/status)
  --version                          Display the version number
  --help                             Show this message

Deprecated:
  -d, --device <index>           Select device by listing index

Copyright 2018–2023 Jonas Malaco, Marshall Asch, CaseySJ, Tom Frey, Andrew
Robertson, ParkerMc, Aleksa Savic, Shady Nawara and contributors

Some modules also incorporate or use as reference work by leaty, Ksenija
Stanojevic, Alexander Tong, Jens Neumaier, Kristóf Jakab, Sean Nelson, Chris
Griffith, notaz, realies and Thomas Pircher. This is mentioned in the module
docstring, along with appropriate additional copyright notices.

SPDX-License-Identifier: GPL-3.0-or-later

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
"""

import datetime
import errno
import inspect
import json
import locale
import logging
import os
import platform
import re
import sys
from numbers import Number
from traceback import format_exception

import colorlog
from docopt import docopt

from liquidctl import __version__
from liquidctl.driver import *
from liquidctl.error import LiquidctlError
from liquidctl.util import color_from_str, fan_mode_parser


# conversion from CLI arg to internal option; as options as forwarded to bused
# and drivers, they must:
#  - have no default value in the CLI level (not forwarded unless explicitly set);
#  - and avoid unintentional conflicts with target function arguments
_PARSE_ARG = {
    '--vendor': lambda x: int(x, 16),
    '--product': lambda x: int(x, 16),
    '--release': lambda x: int(x, 16),
    '--serial': str,
    '--bus': str,
    '--address': str,
    '--usb-port': lambda x: tuple(map(int, x.split('.'))),
    '--match': str,
    '--pick': int,

    '--speed': str.lower,
    '--time-per-color': int,
    '--time-off': int,
    '--alert-threshold': int,
    '--alert-color': color_from_str,
    '--temperature-sensor': int,
    '--direction': str.lower,
    '--start-led': int,
    '--maximum-leds': int,

    '--single-12v-ocp': bool,
    '--pump-mode': str.lower,
    '--legacy-690lc': bool,
    '--non-volatile': bool,
    '--direct-access': bool,
    '--fan-mode': lambda x: fan_mode_parser(x),
    '--unsafe': lambda x: x.lower().split(','),
    '--verbose': bool,
    '--debug': bool,
}

# options that cause liquidctl.driver.find_liquidctl_devices to ommit devices
_FILTER_OPTIONS = [
    'vendor',
    'product',
    'release',
    'serial',
    'bus',
    'address',
    'usb-port',
    'match',
    'pick',
    # --device generates no option
]

# custom number formats for values of select units
_VALUE_FORMATS = {
    '%': '.0f',
    'A': '.2f',
    'V': '.2f',
    'W': '.2f',
    'rpm': '.0f',
    '°C': '.1f',
}

_LOGGER = logging.getLogger(__name__)


def _list_devices_objs(devices):

    def getattr_or(object, name, default=None):
        """Call `getattr` and return `default` on exceptions."""
        try:
            return getattr(object, name, default)
        except Exception:
            return default

    return [
        {
            # replace the experimental suffix with a proper field
            'description': dev.description.replace(' (experimental)', ''),
            'vendor_id': dev.vendor_id,
            'product_id': dev.product_id,
            'release_number': dev.release_number,
            'serial_number': getattr_or(dev, 'serial_number', None),
            'bus': dev.bus,
            'address': dev.address,
            'port': dev.port,
            'driver': type(dev).__name__,
            'experimental': dev.description.endswith('(experimental)'),
        }
        for dev in devices
    ]


def _list_devices_human(devices, *, using_filters, device_id, verbose, debug, **opts):
    for i, dev in enumerate(devices):
        warnings = []

        if not using_filters:
            print(f'Device #{i}: {dev.description}')
        elif device_id is not None:
            print(f'Device #{device_id}: {dev.description}')
        else:
            print(f'Result #{i}: {dev.description}')
        if not verbose:
            continue

        if dev.vendor_id:
            print(f'├── Vendor ID: {dev.vendor_id:#06x}')
        if dev.product_id:
            print(f'├── Product ID: {dev.product_id:#06x}')
        if dev.release_number:
            print(f'├── Release number: {dev.release_number:#06x}')
        try:
            if dev.serial_number:
                print(f'├── Serial number: {dev.serial_number}')
        except:
            msg = 'could not read the serial number'
            if sys.platform.startswith('linux') and os.geteuid:
                msg += ' (requires root privileges)'
            elif sys.platform in ['win32', 'cygwin'] and 'Hid' not in type(dev.device).__name__:
                msg += ' (device possibly requires a kernel driver)'
            if debug:
                _LOGGER.exception(msg.capitalize())
            else:
                warnings.append(msg)

        print(f'├── Bus: {dev.bus}')
        print(f'├── Address: {dev.address}')
        if dev.port:
            port = '.'.join(map(str, dev.port))
            print(f'├── Port: {port}')

        print(f'└── Driver: {type(dev).__name__}')
        if debug:
            driver_hier = (i.__name__ for i in inspect.getmro(type(dev)))
            _LOGGER.debug('MRO: %s', ', '.join(driver_hier))

        for msg in warnings:
            _LOGGER.warning(msg)
        print('')

    assert 'device' not in opts or len(devices) <= 1, 'too many results listed with --device'


def _dev_status_obj(dev, status):
    # don't suppress devices without status data (typically from initialize)
    if not status:
        status = []

    # convert to types that are more suitable to serialization, when that
    # cannot be done later (e.g. because it requires adjusting the `unit`)
    def convert(i):
        key, val, unit = i
        if isinstance(val, datetime.timedelta):
            val = val.total_seconds()
            unit = 's'
        return { 'key': key, 'value': val, 'unit': unit }

    # suppress the experimental suffix, `list` reports it in `.experimental`
    return {
        'bus': dev.bus,
        'address': dev.address,
        'description': dev.description.replace(' (experimental)', ''),
        'status': [convert(x) for x in status]
    }


def _print_dev_status(dev, status):
    if not status:
        return
    print(dev.description)
    tmp = []
    kcols, vcols = 0, 0
    for k, v, u in status:
        if isinstance(v, datetime.timedelta):
            v = str(v)
        elif isinstance(v, bool):
            v = 'Yes' if v else 'No'
        elif v is None:
            v = 'N/A'
        else:
            valfmt = _VALUE_FORMATS.get(u, '')
            v = f'{v:{valfmt}}'
        kcols = max(kcols, len(k))
        vcols = max(vcols, len(v))
        tmp.append((k, v, u))
    for k, v, u in tmp[:-1]:
        print(f'├── {k:<{kcols}}    {v:>{vcols}}  {u}')
    k, v, u = tmp[-1]
    print(f'└── {k:<{kcols}}    {v:>{vcols}}  {u}')
    print('')


def _device_set_color(dev, args, **opts):
    color = map(color_from_str, args['<color>'])
    dev.set_color(args['<channel>'].lower(), args['<mode>'].lower(), color, **opts)

def _device_set_screen(dev, args, **opts):
    dev.set_screen(args["<channel>"], args["<mode>"], args["<value>"], **opts)

def _device_set_speed(dev, args, **opts):
    if len(args['<temperature>']) > 0:
        profile = zip(map(int, args['<temperature>']), map(int, args['<percentage>']))
        dev.set_speed_profile(args['<channel>'].lower(), profile, **opts)
    else:
        dev.set_fixed_speed(args['<channel>'].lower(), int(args['<percentage>'][0]), **opts)


def _make_opts(args):
    opts = {}
    for arg, val in args.items():
        if val is not None and arg in _PARSE_ARG:
            opt = arg.replace('--', '').replace('-', '_')
            opts[opt] = _PARSE_ARG[arg](val)
    return opts


def _log_env_infos():
    _LOGGER.debug('script: %s', sys.argv[0])
    _LOGGER.debug('version: %s', __version__)
    _LOGGER.debug('platform: %s', platform.platform())
    _LOGGER.debug('python: %s', sys.version.replace('\n', ' '))
    _LOGGER.debug('encoding: %s default, %s preferred, utf8_mode %s',
                  locale.getdefaultlocale()[1], locale.getpreferredencoding(), sys.flags.utf8_mode)

    if sys.hexversion >= 0x03080000:
        from importlib.metadata import distribution, version, PackageNotFoundError

        try:
            dist = distribution('liquidctl')
        except PackageNotFoundError:
            _LOGGER.debug('not installed, package metadata not available')
            return

        for req in dist.requires:
            name = re.search('^[a-zA-Z0-9]([a-zA-Z0-9._-]*)', req).group(0)
            try:
                _LOGGER.debug('with %s: %s', name, version(name))
            except Exception as err:
                _LOGGER.debug('with %s: version n/a (%s)', name, err)
    else:
        _LOGGER.debug('importlib.metadata not available')


class _ErrorAcc:
    __slots__ = ['_errors']

    def __init__(self):
        self._errors = 0

    def log(self, msg, *args, err=None, show_err=False):
        self._errors += 1
        if err:
            # log the err with traceback before reporting it properly, this time
            # without traceback; this puts error messages are at the bottom of the
            # output, where most users first look for them
            _LOGGER.info('detailed error: %s: %r', msg, err, *args, exc_info=True)

        if show_err and err:
            _LOGGER.error('%s: %r', msg, err, *args)
        else:
            _LOGGER.error(msg, *args)

    def exit_code(self):
        return 0 if self.is_empty() else 1

    def is_empty(self):
        return not bool(self._errors)


def main():
    args = docopt(__doc__)

    if args['--version']:
        print(f'liquidctl v{__version__} ({platform.platform()})')
        sys.exit(0)

    if args['--debug']:
        args['--verbose'] = True
        log_fmt = '%(log_color)s[%(levelname)s] (%(module)s) (%(funcName)s): %(message)s'
        log_level = logging.DEBUG
    elif args['--verbose']:
        log_fmt = '%(log_color)s%(levelname)s: %(message)s'
        log_level = logging.INFO
    else:
        log_fmt = '%(log_color)s%(levelname)s: %(message)s'
        log_level = logging.WARNING
        sys.tracebacklimit = 0

    if sys.platform == 'win32':
        log_colors = {
            'DEBUG': f'bold_blue',
            'INFO': f'bold_purple',
            'WARNING': 'yellow,bold',
            'ERROR': 'red,bold',
            'CRITICAL': 'red,bold,bg_white',
        }
    else:
        log_colors = {
            'DEBUG': f'blue',
            'INFO': f'purple',
            'WARNING': 'yellow,bold',
            'ERROR': 'red,bold',
            'CRITICAL': 'red,bold,bg_white',
        }

    log_fmtter = colorlog.TTYColoredFormatter(fmt=log_fmt, stream=sys.stderr,
                                              log_colors=log_colors)

    log_handler = logging.StreamHandler()
    log_handler.setFormatter(log_fmtter)
    logging.basicConfig(level=log_level, handlers=[log_handler])

    _log_env_infos()

    if __name__ == '__main__':
        _LOGGER.warning('python -m liquidctl.cli is deprecated, prefer python -m liquidctl')

    errors = _ErrorAcc()

    # unlike humans, machines want to know everything; imply verbose everywhere
    # other than when setting default logging level and format (which are
    # inherently for human consumption)
    if args['--json']:
        args['--verbose'] = True

    opts = _make_opts(args)
    filter_count = sum(1 for opt in opts if opt in _FILTER_OPTIONS)
    device_id = None

    if not args['--device']:
        selected = list(find_liquidctl_devices(**opts))
    else:
        _LOGGER.warning('-d/--device is deprecated, prefer --match or other selection options')
        device_id = int(args['--device'])
        no_filters = {opt: val for opt, val in opts.items() if opt not in _FILTER_OPTIONS}
        compat = list(find_liquidctl_devices(**no_filters))
        if device_id < 0 or device_id >= len(compat):
            errors.log('device index out of bounds')
            return errors.exit_code()
        if filter_count:
            # check that --device matches other filter criteria
            matched_devs = [dev.device for dev in find_liquidctl_devices(**opts)]
            if compat[device_id].device not in matched_devs:
                errors.log('device index does not match remaining selection criteria')
                return errors.exit_code()
            _LOGGER.warning('mixing --device <id> with other filters is not recommended; '
                            'to disambiguate between results prefer --pick <result>')
        selected = [compat[device_id]]

    if args['list']:
        if args['--json']:
            objs = _list_devices_objs(selected)
            print(json.dumps(objs, ensure_ascii=(os.getenv('LANG', None) == 'C')))
        else:
            _list_devices_human(selected, using_filters=bool(filter_count),
                                device_id=device_id, json=json, **opts)
        return

    if len(selected) > 1 and not (args['status'] or args['all']):
        errors.log('too many devices, filter or select one (see: liquidctl --help)')
        return errors.exit_code()
    elif len(selected) == 0:
        errors.log('no device matches available drivers and selection criteria')
        return errors.exit_code()

    # for json
    obj_buf = []

    for dev in selected:
        _LOGGER.debug('device: %s', dev.description)
        try:
            with dev.connect(**opts):
                if args['initialize']:
                    status = dev.initialize(**opts)
                    if args['--json']:
                        obj_buf.append(_dev_status_obj(dev, status))
                    else:
                        _print_dev_status(dev, status)
                elif args['status']:
                    status = dev.get_status(**opts)
                    if args['--json']:
                        obj_buf.append(_dev_status_obj(dev, status))
                    else:
                        _print_dev_status(dev, status)
                elif args['set'] and args['speed']:
                    _device_set_speed(dev, args, **opts)
                elif args['set'] and args['color']:
                    _device_set_color(dev, args, **opts)
                elif args['set'] and args['screen']:
                    _device_set_screen(dev, args, **opts)
                else:
                    assert False, 'unreachable'
        except LiquidctlError as err:
            errors.log(f'{dev.description}: {err}', err=err)
        except OSError as err:
            # each backend API returns a different subtype of OSError (OSError,
            # usb.core.USBError or PermissionError) for permission issues
            if err.errno in [errno.EACCES, errno.EPERM]:
                errors.log(f'{dev.description}: insufficient permissions', err=err)
            elif err.args == ('open failed', ):
                errors.log(
                    f'{dev.description}: could not open, possibly due to insufficient permissions',
                    err=err
                )
            else:
                errors.log(f'{dev.description}: unexpected OS error', err=err, show_err=True)
        except Exception as err:
            errors.log(f'{dev.description}: unexpected error', err=err, show_err=True)

    if errors.is_empty() and args['--json']:
        # use __str__ for values that cannot be directly serialized to JSON
        # (e.g. enums)
        print(json.dumps(obj_buf, ensure_ascii=(os.getenv('LANG', None) == 'C'),
                         default=lambda x: str(x)))

    return errors.exit_code()


if __name__ == '__main__':
    sys.exit(main())



import pytest
from _testutils import MockPyusbDevice, MockRuntimeStorage

from collections import deque

from liquidctl.driver.asetek import Modern690Lc, Legacy690Lc, Hydro690Lc


@pytest.fixture
def mockModern690LcDevice():
    device = MockPyusbDevice()
    dev = Modern690Lc(device, 'Mock Modern 690LC')

    dev.connect()
    return dev


@pytest.fixture
def mockLegacy690LcDevice():
    device = MockPyusbDevice(vendor_id=0xffff, product_id=0xb200, bus=1, port=(1,))
    dev = Legacy690Lc(device, 'Mock Legacy 690LC')

    runtime_storage = MockRuntimeStorage(key_prefixes=['testing'])
    runtime_storage.store('leds_enabled', 0)

    dev.connect(runtime_storage=runtime_storage)
    return dev


def test_modern690Lc_device_not_totally_broken(mockModern690LcDevice):
    """A few reasonable example calls do not raise exceptions."""
    dev = mockModern690LcDevice

    dev.initialize()
    dev.get_status()
    dev.set_color(channel='led', mode='blinking', colors=iter([[3, 2, 1]]),
                  time_per_color=3, time_off=1, alert_threshold=42,
                  alert_color=[90, 80, 10])
    dev.set_color(channel='led', mode='rainbow', colors=[], speed=5)
    dev.set_speed_profile(channel='fan',
                          profile=iter([(20, 20), (30, 50), (40, 100)]))
    dev.set_fixed_speed(channel='pump', duty=50)


def test_modern690Lc_device_connect(mockModern690LcDevice):
    def mock_open():
        nonlocal opened
        opened = True
    mockModern690LcDevice.device.open = mock_open
    opened = False

    with mockModern690LcDevice.connect() as cm:
        assert cm == mockModern690LcDevice
        assert opened


def test_modern690Lc_device_begin_transaction(mockModern690LcDevice):
    mockModern690LcDevice.device._reset_sent()

    mockModern690LcDevice.get_status()

    (begin, _) = mockModern690LcDevice.device._sent_xfers
    xfer_type, bmRequestType, bRequest, wValue, wIndex, datalen = begin
    assert xfer_type == 'ctrl_transfer'
    assert bRequest == 2
    assert bmRequestType == 0x40
    assert wValue == 1
    assert wIndex == 0
    assert datalen is None


def test_modern690Lc_initialize_can_persist_settings(mockModern690LcDevice):
    dev = mockModern690LcDevice
    dev.device._reset_sent()

    dev.initialize(non_volatile=True)

    _begin, (cmd_msgtype, cmd_ep, cmd_data) = dev.device._sent_xfers[-2:]
    assert cmd_msgtype == 'write'
    assert cmd_ep == 2
    assert cmd_data == [0x21]


def test_modern690Lc_set_color_can_persist_settings(mockModern690LcDevice):
    dev = mockModern690LcDevice
    dev.device._reset_sent()

    dev.set_color(channel='led', mode='rainbow', colors=[], non_volatile=True)

    _begin, (cmd_msgtype, cmd_ep, cmd_data) = dev.device._sent_xfers[-2:]
    assert cmd_msgtype == 'write'
    assert cmd_ep == 2
    assert cmd_data == [0x21]


def test_modern690Lc_set_speed_profile_can_persist_settings(mockModern690LcDevice):
    dev = mockModern690LcDevice
    dev.device._reset_sent()

    dev.set_speed_profile(channel='fan', profile=[(20, 20), (30, 50), (40, 100)],
                          non_volatile=True)

    _begin, (cmd_msgtype, cmd_ep, cmd_data) = dev.device._sent_xfers[-2:]
    assert cmd_msgtype == 'write'
    assert cmd_ep == 2
    assert cmd_data == [0x21]


def test_modern690Lc_set_fixed_speed_can_persist_settings(mockModern690LcDevice):
    dev = mockModern690LcDevice
    dev.device._reset_sent()

    dev.set_fixed_speed(channel='pump', duty=50, non_volatile=True)

    _begin, (cmd_msgtype, cmd_ep, cmd_data) = dev.device._sent_xfers[-2:]
    assert cmd_msgtype == 'write'
    assert cmd_ep == 2
    assert cmd_data == [0x21]


def test_legacy690Lc_device_not_totally_broken(mockLegacy690LcDevice):
    """A few reasonable example calls do not raise exceptions."""
    dev = mockLegacy690LcDevice

    dev.initialize()
    status = dev.get_status()
    dev.set_color(channel='led', mode='blinking', colors=iter([[3, 2, 1]]),
                  time_per_color=3, time_off=1, alert_threshold=42,
                  alert_color=[90, 80, 10])
    dev.set_fixed_speed(channel='fan', duty=80)
    dev.set_fixed_speed(channel='pump', duty=50)


def test_legacy690Lc_device_matches_leviathan_updates(mockLegacy690LcDevice):
    dev = mockLegacy690LcDevice

    dev.initialize()
    dev.set_fixed_speed(channel='pump', duty=50)

    dev.device._reset_sent()
    dev.set_color(channel='led', mode='fading', colors=[[0, 0, 255], [0, 255, 0]],
                  time_per_color=1, alert_threshold=60, alert_color=[0, 0, 0])
    _begin, (color_msgtype, color_ep, color_data) = dev.device._sent_xfers
    assert color_msgtype == 'write'
    assert color_ep == 2
    assert color_data[0:12] == [0x10, 0, 0, 255, 0, 255, 0, 0, 0, 0, 0x3c, 1]

    dev.device._reset_sent()
    dev.set_fixed_speed(channel='fan', duty=50)
    _begin, pump_message, fan_message = dev.device._sent_xfers
    assert pump_message == ('write', 2, [0x13, 50])
    assert fan_message == ('write', 2, [0x12, 50])


def test_legacy690Lc_device_initialize_warns_on_non_volatile(mockLegacy690LcDevice, caplog):
    mockLegacy690LcDevice.initialize(non_volatile=True)

    assert 'device does not support non-volatile' in caplog.text


def test_legacy690Lc_device_set_color_warns_on_non_volatile(mockLegacy690LcDevice, caplog):
    mockLegacy690LcDevice.set_color(channel='led', mode='blinking',
                                    colors=iter([[3, 2, 1]]), non_volatile=True)

    assert 'device does not support non-volatile' in caplog.text


def test_legacy690Lc_device_set_fixed_speed_warns_on_non_volatile(mockLegacy690LcDevice, caplog):
    mockLegacy690LcDevice.set_fixed_speed(channel='fan', duty=80, non_volatile=True)

    assert 'device does not support non-volatile' in caplog.text



"""liquidctl driver for ASUS Aura LED USB controllers.

Supported controllers:

- ASUS Aura LED Controller: idVendor=0x0b05, idProduct=0x19af

This controller is found on ASUS ProArt Z690-Creator WiFi and other boards

Additional controllers (requires user testing and feedback):

- idVendor=0x0B05, idProduct=0x1939
- idVendor=0x0B05, idProduct=0x18F3

These controllers have not been sufficiently tested. Users are asked to test and
report issues.

Limitations:

This controller only supports 'effect' modes that are managed entirely by the
controller. The other mode is 'direct' that is managed by software, but this
requires the continuous transmission of command codes to control dynamic lighting
effects (i.e. all lighting modes in which colors either change or the LED 
blinks/fades/pulses).

In 'effect' mode, the software simply issues a  'fire-and-forget' command to the
controller, which subsequently manages the lighting effect on its own. There are
some limitations in 'effect' mode, as follows:

- Dynamic color modes (spectrum, rainbow, etc.) cannot be applied to individual
  color channels, but apply to all channels

- Static color modes (static) can be applied to individual color channels

- Off mode turns all channels off regardless of which channel is selected. However,
  individual channels can subsequently be enabled, while others will remain off.
  This applies only to static modes.


Acknowledgements:

- Aura Addressable Header Controller (for list of color mode names)
  https://gitlab.com/cneil02/aura-addressable-header-controller

- OpenRGB Project (for list of color mode names)
  https://github.com/CalcProgrammer1/OpenRGB

- @dehjomz for discovering color modes 0x10, 0x11, 0x12, 0x13

Aura LED control codes were independently obtained from USB traffic captured using
Wireshark on Windows. This Aura LED controller uses very different control codes
from previous Aura LED controllers.

Copyright CaseySJ and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import logging
import sys
from collections import namedtuple

from liquidctl.driver.usb import UsbHidDriver
from liquidctl.error import NotSupportedByDevice
from liquidctl.util import clamp

_LOGGER = logging.getLogger(__name__)

_READ_LENGTH = 65
_WRITE_LENGTH = 65

_RESET_CHANNEL_ITERATIONS = 5  # Aura Crate software uses 38

_CMD_CODE = 0xEC

_FUNCTION_CODE = {
    "direct": [_CMD_CODE, 0x40],
    "end_direct": [_CMD_CODE, 0x35, 0x00, 0x00, 0x00, 0x00],
    "end_effect": [_CMD_CODE, 0x35, 0x00, 0x00, 0x00, 0xFF],
    "start_seq1": [_CMD_CODE, 0x35],
    "start_seq2": [_CMD_CODE, 0x36, 0x00],
    "end_seq1": [_CMD_CODE, 0x35, 0x00, 0x00, 0x01, 0x05],
    "end_seq2": [_CMD_CODE, 0x3F, 0x55, 0x00, 0x00],
    "reset_seq1": [_CMD_CODE, 0x38, 0x01, 0x01],
    "reset_seq2": [_CMD_CODE, 0x38, 0x00, 0x01],
    "channel_off_pre1": [_CMD_CODE, 0x38, 0x01, 0x00],
    "channel_off_pre2": [_CMD_CODE, 0x38, 0x00, 0x00],
    "channel_off_prefix": [_CMD_CODE, 0x35, 0x01, 0x00, 0x00, 0xFF],
    "firmware": [_CMD_CODE, 0x82],
    "config": [_CMD_CODE, 0xB0],
}

# channel_type 0 designates RGB bus
# channel_type 1 designates ARGB bus
# "effect" mode channel IDs are different from "direct" mode channel IDs
_ColorChannel = namedtuple(
    "_ColorChannel", ["name", "channel_id", "direct_channel_id", "channel_type", "rgb_offset"]
)

_COLOR_CHANNELS = {
    channel.name: channel
    for channel in [
        _ColorChannel("led1", 0x01, 0x00, 0x00, 0x0),  # rgb channel
        _ColorChannel("led2", 0x02, 0x00, 0x01, 0x01),  # argb channel
        _ColorChannel("led3", 0x04, 0x01, 0x01, 0x02),  # argb channel
        _ColorChannel("led4", 0x08, 0x02, 0x01, 0x03),  # argb channel
    ]
}

_ColorMode = namedtuple("_ColorMode", ["name", "value", "takes_color"])

_COLOR_MODES = {
    mode.name: mode
    for mode in [
        _ColorMode("off", 0x00, takes_color=False),
        _ColorMode("static", 0x01, takes_color=True),
        _ColorMode("breathing", 0x02, takes_color=True),
        _ColorMode("flashing", 0x03, takes_color=True),
        _ColorMode("spectrum_cycle", 0x04, takes_color=False),
        _ColorMode("rainbow", 0x05, takes_color=False),
        _ColorMode("spectrum_cycle_breathing", 0x06, takes_color=False),
        _ColorMode("chase_fade", 0x07, takes_color=True),
        _ColorMode("spectrum_cycle_chase_fade", 0x08, takes_color=False),
        _ColorMode("chase", 0x09, takes_color=True),
        _ColorMode("spectrum_cycle_chase", 0x0A, takes_color=False),
        _ColorMode("spectrum_cycle_wave", 0x0B, takes_color=False),
        _ColorMode("chase_rainbow_pulse", 0x0C, takes_color=False),
        _ColorMode("rainbow_flicker", 0x0D, takes_color=False),
        _ColorMode("gentle_transition", 0x10, takes_color=False),
        _ColorMode("wave_propagation", 0x11, takes_color=False),
        _ColorMode("wave_propagation_pause", 0x12, takes_color=False),
        _ColorMode("red_pulse", 0x13, takes_color=False),
    ]
}


class AuraLed(UsbHidDriver):
    """
    liquidctl driver for ASUS Aura LED USB controllers.
    This driver only supports 'effect' mode, hence no speed/color channels

    Devices 0x1939 and 0x18F3 are not fully supported at this time; users are asked
    to experiment with this driver and provide feedback
    """

    _MATCHES = [
        (0x0B05, 0x19AF, "ASUS Aura LED Controller (experimental)", {}),
        (0x0B05, 0x1939, "ASUS Aura LED Controller (experimental)", {}),
        (0x0B05, 0x18F3, "ASUS Aura LED Controller (experimental)", {}),
    ]

    def initialize(self, **kwargs):
        """Initialize the device.
        Returns a list of `(property, value, unit)` tuples, containing the
        firmware version and other useful information provided by the hardware.
        """
        # Get firmware version
        self._write(_FUNCTION_CODE["firmware"])

        # Build reply string
        status = []
        data = self.device.read(_READ_LENGTH)
        if data[1] == 0x02:
            status.append(("Firmware version", "".join(map(chr, data[2:17])), ""))
        else:
            status.append("Unexpected reply for firmware", "", "")
            return status

        # This stops Direct mode if it was previously applied
        self._write(_FUNCTION_CODE["end_direct"])

        """
        Extra operations during initialization
        This is experimental and may not be necessary
        self._write([0xec, 0x31, 0x0d, 0x00]);
        self._write([0xec, 0xb1, 0x00, 0x00]);
        self.device.read(_READ_LENGTH)
        self._write([0xec, 0x31, 0x0e, 0x00]);
        self._write([0xec, 0xb1, 0x00, 0x00]);
        self.device.read(_READ_LENGTH)
        self._write([0xec, 0x31, 0x0f, 0x00]);
        self._write([0xec, 0xb1, 0x00, 0x00]);
        self.device.read(_READ_LENGTH)
        self._write([0xec, 0x83, 0x00, 0x00]);
        self.device.read(_READ_LENGTH)
        """
        return status

    def get_status(self, **kwargs):
        """Get a status report."""
        status = []
        # Get config table
        self._write(_FUNCTION_CODE["config"])
        data = self.device.read(_READ_LENGTH)
        if data[1] == 0x30:
            start_index = 4  # index of first record

            argb_channels = data[start_index + 2]
            rgb_channels = data[start_index + 3]
            status.append(("ARGB channels: " + str(argb_channels), "", ""))
            status.append((" RGB channels: " + str(rgb_channels), "", ""))

            if "debug" in kwargs and kwargs["debug"] == True:
                num = 6  # number of bytes per record
                count = 1
                while start_index + num < _READ_LENGTH:
                    status.append(
                        (
                            "Device Config: " + str(count),
                            ", ".join(
                                "0x{:02x}".format(x) for x in data[start_index : start_index + num]
                            ),
                            "",
                        )
                    )
                    start_index += num
                    count += 1
        else:
            status.append("Unexpected reply for config", "", "")
        return status

    def set_color(self, channel, mode, colors, speed="normal", **kwargs):
        """Set the color mode for a specific channel.

        `colors` should be an iterable of zero or one `[red, green, blue]`
        triples, where each red/green/blue component is a value in the range
        0–255.
        """
        colors = iter(colors)
        if _COLOR_MODES[mode].takes_color:
            try:
                r, g, b = next(colors)
                single_color = (r, g, b)
            except:
                raise ValueError(f"one color required for this mode") from None
        else:
            single_color = (0, 0, 0)

        if channel != "sync" and channel not in _COLOR_CHANNELS:
            message = "valid channels are "
            for chan in _COLOR_CHANNELS:
                message += chan + " "
            raise KeyError(message) from None
            return

        """
        This is experimental (it's an example of direct mode)
        if mode == 'off':
            self.channel_off(channel)
            self.reset_all_channels()
            return
        """

        if channel == "sync":
            selected_channels = _COLOR_CHANNELS.values()
        else:
            selected_channels = (_COLOR_CHANNELS[channel],)
        full_cmd_seq = []  # entire series of commands are added to this list

        """
        Experimental code for treating RGB channel differently from others
        if channel == "led1":
            cmd_tuple=self.construct_color_commands(channel, mode, single_color)
            self._write(cmd_tuple[0])
            self._write(cmd_tuple[1])
            self._write(_FUNCTION_CODE["end_seq2"])
            self.end_color_sequence()
            self._write(cmd_tuple[0])
            self._write(cmd_tuple[1])
            self._write(_FUNCTION_CODE["end_seq2"])
        else:
        """

        for chan in selected_channels:
            cmd_tuple = self.construct_color_commands(chan.name, mode, single_color)
            full_cmd_seq.append(cmd_tuple[0])
            full_cmd_seq.append(cmd_tuple[1])
            full_cmd_seq.append(_FUNCTION_CODE["end_seq2"])

            """
            ASUS Aura Crate sends command sequence twice, but our tests show
            that this may be redundant. Nevertheless, let's keep this code here
            in case we need to send commands twice as well
            #full_cmd_seq.append(cmd_tuple[0])
            #full_cmd_seq.append(cmd_tuple[1])
            #full_cmd_seq.append(_FUNCTION_CODE["end_seq2"])
            """

        for cmd_seq in full_cmd_seq:
            self._write(cmd_seq)
        self.end_color_sequence()

    def set_speed_profile(self, channel, profile, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()

    def set_fixed_speed(self, channel, duty, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()

    def reset_all_channels(self):
        """Reset all LED channels."""
        for i in range(_RESET_CHANNEL_ITERATIONS):
            self._write(_FUNCTION_CODE["reset_seq1"])
            self._write(_FUNCTION_CODE["reset_seq2"])

    def channel_off(self, channel):
        """
        Uses direct mode to disable a specific channel
        """
        self._write(_FUNCTION_CODE["end_effect"])

        for i in range(_RESET_CHANNEL_ITERATIONS):
            self._write(_FUNCTION_CODE["channel_off_pre1"])
            self._write(_FUNCTION_CODE["channel_off_pre2"])

        self._write(_FUNCTION_CODE["channel_off_prefix"])

        # set all LEDs to off, 20 at a time
        for i in (0, 20, 40, 60, 80, 100):
            self._write(
                _FUNCTION_CODE["direct"]
                + [_COLOR_CHANNELS[channel].direct_channel_id | (0x80 * (i == 100)), i, 20]
            )
        self.end_color_sequence()
        self._write(_FUNCTION_CODE["end_direct"])

    def construct_color_commands(self, channel, mode, single_color):
        """
        Create command strings for specified color channel
        """
        mode = _COLOR_MODES[mode]
        channel_type = _COLOR_CHANNELS[channel].channel_type  # 0=RGB, 1=ARGB
        channel_id = _COLOR_CHANNELS[channel].channel_id
        rgb_offset = _COLOR_CHANNELS[channel].rgb_offset
        data1 = _FUNCTION_CODE["start_seq1"] + [channel_type, 0x00, 0x00, mode.value]
        data2 = _FUNCTION_CODE["start_seq2"] + [channel_id, 0x00] + [0, 0, 0] * rgb_offset
        data2 += single_color
        return (data1, data2)

    def end_color_sequence(self):
        self._write(_FUNCTION_CODE["end_seq1"])
        self._write(_FUNCTION_CODE["end_seq2"])

    def _write(self, data):
        padding = [0x0] * (_WRITE_LENGTH - len(data))
        self.device.write(data + padding)

    def set_screen(self, channel, mode, value, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()



import pytest
from _testutils import MockHidapiDevice, Report

from liquidctl.driver.nzxt_epsu import NzxtEPsu


class _MockPsuDevice(MockHidapiDevice):
    def write(self, data):
        super().write(data)
        data = data[1:]  # skip unused report ID
        reply = bytearray(64)
        reply[0:2] = (0xaa, data[2])
        if data[5] == 0x06:
            reply[2] = data[2] - 2
        elif data[5] == 0xfc:
            reply[2:4] = (0x11, 0x41)
        self.preload_read(Report(0, reply[0:]))


@pytest.fixture
def mockPsuDevice():
    device = _MockPsuDevice()
    return NzxtEPsu(device, 'mock NZXT E500 PSU')


def test_psu_device_initialize(mockPsuDevice):
    mockPsuDevice.initialize()

    assert len(mockPsuDevice.device.sent) == 0


def test_psu_device_status(mockPsuDevice):

    mockPsuDevice.connect()
    status = mockPsuDevice.get_status()

    fw = next(filter(lambda x: x[0] == 'Firmware version', status))
    assert fw == ('Firmware version', 'A017/40983', '')

    sent = mockPsuDevice.device.sent
    assert sent[0] == Report(0, [0xad, 0, 3, 1, 0x60, 0xfc] + 58*[0])



import pytest
from _testutils import VirtualSmbus

from liquidctl.driver.nvidia import *
from liquidctl.error import *


# EVGA Pascal


@pytest.fixture
def evga_1080_ftw_bus():
    return VirtualSmbus(
            description='NVIDIA i2c adapter 1 at 1:00.0',
            parent_vendor=NVIDIA,
            parent_device=NVIDIA_GTX_1080,
            parent_subsystem_vendor=EVGA,
            parent_subsystem_device=EVGA_GTX_1080_FTW,
            parent_driver='nvidia',
    )


def test_evga_pascal_finds_devices(evga_1080_ftw_bus, monkeypatch):
    smbus = evga_1080_ftw_bus

    checks = [
        ('parent_subsystem_vendor', 0xffff),
        ('parent_vendor', 0xffff),
        ('parent_driver', 'other'),
        ('parent_subsystem_device', 0xffff),
        ('parent_device', 0xffff),
        ('description', 'NVIDIA i2c adapter 2 at 1:00.0'),
    ]

    for attr, val in checks:
        with monkeypatch.context() as m:
            m.setattr(smbus, attr, val)
            assert list(EvgaPascal.probe(smbus)) == [], \
                f'changing {attr} did not cause a mismatch'

    assert list(map(type, EvgaPascal.probe(smbus))) == [EvgaPascal]


def test_evga_pascal_get_status_is_noop(evga_1080_ftw_bus):
    card = next(EvgaPascal.probe(evga_1080_ftw_bus))
    assert card.get_status() == []


def test_evga_pascal_get_verbose_status_is_unsafe(evga_1080_ftw_bus):
    card = next(EvgaPascal.probe(evga_1080_ftw_bus))
    assert card.get_status(verbose=True) == []
    assert card.get_status(verbose=True, unsafe='other') == []


def test_evga_pascal_gets_verbose_status(evga_1080_ftw_bus):
    enable = ['smbus']
    card = next(EvgaPascal.probe(evga_1080_ftw_bus))

    with card.connect(unsafe=enable):
        evga_1080_ftw_bus.write_byte_data(0x49, 0x09, 0xaa)
        evga_1080_ftw_bus.write_byte_data(0x49, 0x0a, 0xbb)
        evga_1080_ftw_bus.write_byte_data(0x49, 0x0b, 0xcc)
        evga_1080_ftw_bus.write_byte_data(0x49, 0x0c, 0x01)

        status = card.get_status(verbose=True, unsafe=enable)
        expected = [
            ('Mode', EvgaPascal.Mode.FIXED, ''),
            ('Color', 'aabbcc', ''),
        ]

        assert status == expected


def test_evga_pascal_set_color_is_unsafe(evga_1080_ftw_bus):
    card = next(EvgaPascal.probe(evga_1080_ftw_bus))

    with pytest.raises(UnsafeFeaturesNotEnabled):
        card.set_color('led', 'off', [])

    with pytest.raises(UnsafeFeaturesNotEnabled):
        card.set_color('led', 'off', [], unsafe='other')


def test_evga_pascal_sets_color_to_off(evga_1080_ftw_bus):
    enable = ['smbus']
    card = next(EvgaPascal.probe(evga_1080_ftw_bus))

    with card.connect(unsafe=enable):
        # change mode register to something other than 0 (=off)
        evga_1080_ftw_bus.write_byte_data(0x49, 0x0c, 0x01)

        card.set_color('led', 'off', [], unsafe=enable)
        assert evga_1080_ftw_bus.read_byte_data(0x49, 0x0c) == 0x00

        # persistence not enabled
        assert evga_1080_ftw_bus.read_byte_data(0x49, 0x23) == 0x00


def test_evga_pascal_sets_color_to_fixed(evga_1080_ftw_bus):
    enable = ['smbus']
    card = next(EvgaPascal.probe(evga_1080_ftw_bus))

    with card.connect(unsafe=enable):
        radical_red = [0xff, 0x35, 0x5e]
        card.set_color('led', 'fixed', [radical_red], unsafe=enable)

        assert evga_1080_ftw_bus.read_byte_data(0x49, 0x0c) == 0x01
        assert evga_1080_ftw_bus.read_byte_data(0x49, 0x09) == 0xff
        assert evga_1080_ftw_bus.read_byte_data(0x49, 0x0a) == 0x35
        assert evga_1080_ftw_bus.read_byte_data(0x49, 0x0b) == 0x5e

        # persistence not enabled
        assert evga_1080_ftw_bus.read_byte_data(0x49, 0x23) == 0x00


def test_evga_pascal_sets_color_to_rainbow(evga_1080_ftw_bus):
    enable = ['smbus']
    card = next(EvgaPascal.probe(evga_1080_ftw_bus))

    with card.connect(unsafe=enable):
        card.set_color('led', 'rainbow', [], unsafe=enable)
        assert evga_1080_ftw_bus.read_byte_data(0x49, 0x0c) == 0x02

        # persistence not enabled
        assert evga_1080_ftw_bus.read_byte_data(0x49, 0x23) == 0x00


def test_evga_pascal_sets_color_to_breathing(evga_1080_ftw_bus):
    enable = ['smbus']
    card = next(EvgaPascal.probe(evga_1080_ftw_bus))

    with card.connect(unsafe=enable):
        radical_red = [0xff, 0x35, 0x5e]
        card.set_color('led', 'breathing', [radical_red], unsafe=enable)

        assert evga_1080_ftw_bus.read_byte_data(0x49, 0x0c) == 0x05
        assert evga_1080_ftw_bus.read_byte_data(0x49, 0x09) == 0xff
        assert evga_1080_ftw_bus.read_byte_data(0x49, 0x0a) == 0x35
        assert evga_1080_ftw_bus.read_byte_data(0x49, 0x0b) == 0x5e

        # persistence not enabled
        assert evga_1080_ftw_bus.read_byte_data(0x49, 0x23) == 0x00


def test_evga_pascal_sets_non_volatile_color(evga_1080_ftw_bus):
    enable = ['smbus']
    card = next(EvgaPascal.probe(evga_1080_ftw_bus))

    orig = evga_1080_ftw_bus.write_byte_data

    def raise_if_23h(address, register, value):
        orig(address, register, value)
        if register == 0x23:
            raise OSError()  # mimic smbus.SMBus on actual hardware

    evga_1080_ftw_bus.write_byte_data = raise_if_23h

    with card.connect(unsafe=enable):
        card.set_color('led', 'off', [], non_volatile=True, unsafe=enable)
        assert evga_1080_ftw_bus.read_byte_data(0x49, 0x23) == 0xe5


def test_evga_pascal_experimental_devices_are_unsafe():
    for dev_id, sub_dev_id, desc in EvgaPascal._MATCHES:
        if 'experimental' not in desc:
            continue

        vbus = VirtualSmbus(
                description='NVIDIA i2c adapter 1 at 1:00.0',
                parent_vendor=NVIDIA,
                parent_device=dev_id,
                parent_subsystem_vendor=EVGA,
                parent_subsystem_device=sub_dev_id,
                parent_driver='nvidia',
        )

        card = next(EvgaPascal.probe(vbus))

        insufficient = ['smbus']

        # can connect but not read status or set color without specific
        # experimental feature
        with card.connect(unsafe=insufficient):
            assert card.get_status(verbose=True, unsafe=insufficient) == []

            with pytest.raises(UnsafeFeaturesNotEnabled):
                card.set_color('led', 'off', [], unsafe=insufficient)


# ASUS Turing


def create_strix_2080ti_oc_bus(controller_address=None):
    smbus = VirtualSmbus(
            description='NVIDIA i2c adapter 1 at 1c:00.0',
            parent_vendor=NVIDIA,
            parent_device=NVIDIA_RTX_2080_TI_REV_A,
            parent_subsystem_vendor=ASUS,
            parent_subsystem_device=ASUS_STRIX_RTX_2080_TI_OC,
            parent_driver='nvidia',
    )

    if controller_address is None:
        return smbus

    smbus.open()
    try:
        smbus.write_byte_data(controller_address, 0x20, 0x15)
        smbus.write_byte_data(controller_address, 0x21, 0x89)
    except:
        pass
    smbus.close()

    return smbus


@pytest.fixture
def strix_2080ti_oc_bus():
    address = 0x2a  # not the first candidate
    return create_strix_2080ti_oc_bus(address)


def test_rog_turing_does_not_find_devices(monkeypatch):
    smbus = create_strix_2080ti_oc_bus()  # no addresses enabled

    checks = [
        ('parent_subsystem_vendor', 0xffff),
        ('parent_vendor', 0xffff),
        ('parent_driver', 'other'),
        ('parent_subsystem_device', 0xffff),
        ('parent_device', 0xffff),
        ('description', 'NVIDIA i2c adapter 2 at 1:00.0'),
    ]

    for attr, val in checks:
        with monkeypatch.context() as m:
            m.setattr(smbus, attr, val)
            assert list(RogTuring.probe(smbus)) == [], \
                f'changing {attr} did not cause a mismatch'

    # with unsafe features addresses can be checked and none match
    assert list(map(type, RogTuring.probe(smbus, unsafe='smbus'))) == []


def test_rog_turing_assumes_device_if_unsafe_bus_unavailable(monkeypatch):
    smbus = create_strix_2080ti_oc_bus()  # no addresses enabled

    # assume a device if the bus cannot be read due to missing unsafe features
    assert list(map(type, RogTuring.probe(smbus))) == [RogTuring]


def test_rog_turing_finds_devices_on_any_addresses(monkeypatch):
    addresses = [0x29, 0x2a, 0x60]
    for addr in addresses:
        smbus = create_strix_2080ti_oc_bus(controller_address=addr)
        cards = list(RogTuring.probe(smbus, unsafe='smbus'))
        assert list(map(type, cards)) == [RogTuring]
        assert cards[0].address == hex(addr)


def test_rog_turing_only_use_one_address(monkeypatch):
    smbus = create_strix_2080ti_oc_bus()  # no addresses enabled
    addresses = [0x29, 0x2a, 0x60]

    smbus.open()
    for addr in addresses:
        smbus.write_byte_data(addr, 0x20, 0x15)
        smbus.write_byte_data(addr, 0x21, 0x89)
    smbus.close()

    cards = list(RogTuring.probe(smbus, unsafe='smbus'))
    assert list(map(type, cards)) == [RogTuring]
    assert cards[0].address == hex(addresses[0])


def test_rog_turing_unsafely_probed_is_not_usable(strix_2080ti_oc_bus):
    card = next(RogTuring.probe(strix_2080ti_oc_bus))
    too_late = 'smbus'

    with pytest.raises(AssertionError):
        card.get_status(verbose=True, unsafe=too_late)

    with pytest.raises(AssertionError):
        card.set_color('led', 'off', [], unsafe=too_late)


def test_rog_turing_get_status_is_noop(strix_2080ti_oc_bus):
    card = next(RogTuring.probe(strix_2080ti_oc_bus))
    assert card.get_status() == []


def test_rog_turing_get_verbose_status_is_unsafe(strix_2080ti_oc_bus):
    card = next(RogTuring.probe(strix_2080ti_oc_bus))
    assert card.get_status(verbose=True) == []


def test_rog_turing_gets_verbose_status(strix_2080ti_oc_bus):
    enable = ['smbus']
    card = next(RogTuring.probe(strix_2080ti_oc_bus, unsafe=enable))

    with card.connect(unsafe=enable):
        strix_2080ti_oc_bus.write_byte_data(0x2a, 0x07, 0x01)
        strix_2080ti_oc_bus.write_byte_data(0x2a, 0x04, 0xaa)
        strix_2080ti_oc_bus.write_byte_data(0x2a, 0x05, 0xbb)
        strix_2080ti_oc_bus.write_byte_data(0x2a, 0x06, 0xcc)

        status = card.get_status(verbose=True, unsafe=enable)
        expected = [
            ('Mode', RogTuring.Mode.FIXED, ''),
            ('Color', 'aabbcc', ''),
        ]

        assert status == expected


def test_rog_turing_set_color_is_unsafe(strix_2080ti_oc_bus):
    card = next(RogTuring.probe(strix_2080ti_oc_bus))

    with pytest.raises(UnsafeFeaturesNotEnabled):
        assert card.set_color('led', 'off', [])


def test_rog_turing_sets_color_to_off(strix_2080ti_oc_bus):
    enable = ['smbus']
    card = next(RogTuring.probe(strix_2080ti_oc_bus, unsafe=enable))

    with card.connect(unsafe=enable):
        # change colors to something other than 0
        strix_2080ti_oc_bus.write_byte_data(0x2a, 0x04, 0xaa)
        strix_2080ti_oc_bus.write_byte_data(0x2a, 0x05, 0xbb)
        strix_2080ti_oc_bus.write_byte_data(0x2a, 0x06, 0xcc)

        card.set_color('led', 'off', [], unsafe=enable)

        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x07) == 0x01
        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x04) == 0x00
        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x05) == 0x00
        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x06) == 0x00

        # persistence not enabled
        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x0e) == 0x00


def test_rog_turing_sets_color_to_fixed(strix_2080ti_oc_bus):
    enable = ['smbus']
    card = next(RogTuring.probe(strix_2080ti_oc_bus, unsafe=enable))

    with card.connect(unsafe=enable):
        radical_red = [0xff, 0x35, 0x5e]
        card.set_color('led', 'fixed', [radical_red], unsafe=enable)

        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x07) == 0x01
        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x04) == 0xff
        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x05) == 0x35
        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x06) == 0x5e

        # persistence not enabled
        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x0e) == 0x00


def test_rog_turing_sets_color_to_rainbow(strix_2080ti_oc_bus):
    enable = ['smbus']
    card = next(RogTuring.probe(strix_2080ti_oc_bus, unsafe=enable))

    with card.connect(unsafe=enable):
        card.set_color('led', 'rainbow', [], unsafe=enable)
        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x07) == 0x04

        # persistence not enabled
        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x0e) == 0x00


def test_rog_turing_sets_color_to_breathing(strix_2080ti_oc_bus):
    enable = ['smbus']
    card = next(RogTuring.probe(strix_2080ti_oc_bus, unsafe=enable))

    with card.connect(unsafe=enable):
        radical_red = [0xff, 0x35, 0x5e]
        card.set_color('led', 'breathing', [radical_red], unsafe=enable)

        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x07) == 0x02
        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x04) == 0xff
        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x05) == 0x35
        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x06) == 0x5e

        # persistence not enabled
        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x0e) == 0x00


def test_rog_turing_sets_non_volatile_color(strix_2080ti_oc_bus):
    enable = ['smbus']
    card = next(RogTuring.probe(strix_2080ti_oc_bus, unsafe=enable))

    with card.connect(unsafe=enable):
        card.set_color('led', 'off', [], non_volatile=True, unsafe=enable)
        assert strix_2080ti_oc_bus.read_byte_data(0x2a, 0x0e) == 0x01


def test_rog_turing_experimental_devices_are_unsafe():
    for dev_id, sub_dev_id, desc in RogTuring._MATCHES:
        if 'experimental' not in desc:
            continue

        vbus = VirtualSmbus(
                description='NVIDIA i2c adapter 1 at 1:00.0',
                parent_vendor=NVIDIA,
                parent_device=dev_id,
                parent_subsystem_vendor=ASUS,
                parent_subsystem_device=sub_dev_id,
                parent_driver='nvidia',
        )

        card = next(RogTuring.probe(vbus))

        enable = ['smbus', 'experimental_asus_gpu']
        insufficient = ['smbus']

        # not usable if connected without specific experimental feature
        with card.connect(unsafe=insufficient):
            with pytest.raises(AssertionError):
                card.get_status(verbose=True, unsafe=enable)

            with pytest.raises(AssertionError):
                card.set_color('led', 'off', [], unsafe=enable)

        # once connected, still cannot read status or set color without
        # specific experimental feature
        with card.connect(unsafe=enable):
            assert card.get_status(verbose=True, unsafe=insufficient) == []

            with pytest.raises(UnsafeFeaturesNotEnabled):
                card.set_color('led', 'off', [], unsafe=insufficient)
